unit Character;

interface

uses
  World, Tiles;

const
  //Image frame consts
  FrameStanding = 0;
  FrameGoing = 1;
  FrameJumping = 2;
  FrameFalling = 3;
  FrameClimbing = 4;

  //Parameter consts
  GroundVelIncrement: Extended = 0.5; //Inkrement za Vx trcanje
  AirVelIncrement: Extended = 0.125; //Inkrement za Vx letenje
  WaterVelIncrement: Extended = 0.0625; //Inkrement za Vx ronjenje

  GroundSmallMaxVel: Extended = 4; //Max Vx kada hoda
  GroundLargeMaxVel: Extended = 7.5; //Max Vx kada trci
  WaterMaxVel: Extended = 2;//Max Vx kada roni
  AirMaxVel: Extended = 9; //Max Vx kada leti

  GroundTurnDelay: Extended = 3;
  WaterTurnDelay: Extended = 2;
  AirTurnDelay: Extended = 3;

  GroundFriction: Extended = 0.5; //Trenje kada stane sa trcanjem
  AirFriction: Extended = 0.0625; //Trenje kada stane sa letenjem
  WaterFriction: Extended = 0.03125; //Trenje kada stane sa ronjenjem

  GroundSmallJumpImpulse: Extended = 13.75; //Vertikalni impuls malog skoka
  GroundLargeJumpImpulse: Extended = 15.75; //Vertikalni impuls velikog skoka
  WaterJumpImpulse: Extended = 8; //Vertikalni impuls u vodi
  AirJumpImpulse: Extended = 10; //Vertikalni impuls u vazduhu

  GroundGravity: Extended = 0.5; //Gravitacija na zemlji
  AirGravity: Extended = 0.125; //Gravitacija letenja
  WaterGravity: Extended = -0.03125; //Gravitacija ronjenja
type
  TCharacter = class
  private
    FWorld: TWorld;
    FGravity: Extended;
    FFriction: Extended;
    FVelIncrement: Extended;
    FMaxVel: Extended;
    FTurnDelay: Extended;
    FJumpImpulse: Extended;
    FFlags: set of (flgGoingLeft, flgGoingRight, flgJumping, flgRunning, flgFalling, flgGoingUp, flgGoingDown, flgFacingLeft, flgFacingRight);
    FEnvFlags: set of (flgWater, flgAir, flgGround, flgLadder);
    FSpecialUltraMegaDirtyWaterJump: Boolean;
    FWidth: Integer;
    FHeight: Integer;
    FPriorX: Extended;
    FPriorY: Extended;
    FFrameCounter: Byte;
    FFrameSwitch: Boolean;
    SpecialTop, SpecialBtm: Byte;
    procedure ParamCheck;
  public
    FX: Extended;
    FY: Extended;
    FVx: Extended;
    FVy: Extended;
    constructor Create(AWorld: TWorld; AWidth, AHeight, StartX, StartY: Integer);
    procedure Tick;
    procedure GoLeft;
    procedure StopLeft;
    procedure GoRight;
    procedure StopRight;
    procedure GoJump;
    procedure StopJump;
    procedure GoRun;
    procedure StopRun;
    procedure GoDown;
    procedure StopDown;
    procedure GoUp;
    procedure StopUp;
    function X: Integer;
    function Y: Integer;
    function Vx: Extended;
    function Vy: Extended;
    function GetFrameIndex: Byte;
    function IsAlive: Boolean;
  end;

implementation

{ TCharacter }

constructor TCharacter.Create(AWorld: TWorld; AWidth, AHeight, StartX, StartY: Integer);
begin
  FWorld := AWorld; //Potrebno zbog World.IsAvailable metoda
  FX := StartX; //PocetniX
  FY := StartY; //PocetniY
  FWidth := AWidth;
  FHeight := AHeight;
  FVx := 0;
  FVy := 0;
  FFlags := [flgFacingRight];
  FEnvFlags := [flgGround];
  ParamCheck;
end;

procedure TCharacter.GoLeft;
begin
  FFlags := FFlags + [flgGoingLeft];
  FFlags := FFlags + [flgFacingLeft];
  FFlags := FFlags - [flgFacingRight];
  if not (flgFalling in FFlags) then
    if FVx > FTurnDelay then
      FVx := FTurnDelay;
end;

procedure TCharacter.GoRight;
begin
  FFlags := FFlags - [flgGoingLeft];
  FFlags := FFlags + [flgGoingRight];
  FFlags := FFlags - [flgFacingLeft];
  FFlags := FFlags + [flgFacingRight];
  if not (flgFalling in FFlags) then
    if FVx < -FTurnDelay then
      FVx := -FTurnDelay;
end;

procedure TCharacter.GoJump;
begin
  if flgGround in FEnvFlags then
    if not (flgFalling in FFlags) then
    begin
      FVy := FJumpImpulse;
      FFlags := FFlags + [flgJumping];
    end;

  if FSpecialUltraMegaDirtyWaterJump then
  begin
    FVy := WaterJumpImpulse;
     if FVy > GroundLargeJumpImpulse then
       FVy := GroundLargeJumpImpulse;
    FFlags := FFlags + [flgJumping];
    Exit;
  end;
end;

procedure TCharacter.GoRun;
begin
  FFlags := FFlags + [flgRunning];
end;

procedure TCharacter.StopJump;
begin
  if flgJumping in FFlags then
    if flgGround in FEnvFlags then
    begin
      FFlags := FFlags - [flgJumping];
      if FVy > 6 then  //Da ne bi TRENUTNO stao sa skokom pri pustanju Jump tastera
        FVy := 6;
    end;
  FFlags := FFlags - [flgGoingUp];
end;

procedure TCharacter.StopLeft;
begin
  FFlags := FFlags - [flgGoingLeft];
end;

procedure TCharacter.StopRight;
begin
  FFlags := FFlags - [flgGoingRight];
end;

procedure TCharacter.StopRun;
begin
  FFlags := FFlags - [flgRunning];
end;

procedure TCharacter.Tick;
begin
//Osnovne Provere
  ParamCheck;

//Izracunavanje Vx i X
  if flgGoingRight in FFlags then
    FVx := FVx + FVelIncrement;

  if flgGoingLeft in FFlags then
    FVx := FVx - FVelIncrement;

  if FVx > FMaxVel then
    FVx := FMaxVel;
  if FVx < -FMaxVel then
    FVx := -FMaxVel;

  //Ako ne ide ni levo ni desno onda deluje trenje
  if not ((flgGoingLeft in FFlags) or (flgGoingRight in FFlags)) then
    if FVx <> 0 then
      if FVx > 0 then
        FVx := FVx - FFriction
      else
        FVx := FVx + FFriction;

  FPriorX := FX; //Zapamti stari (dobri) X
  FX := FX + FVx; //Dodaj sracunatu brzinu

  //Proveri i eventualno ispravi novi polozaj
  if not (FWorld.IsAvailablePos(Trunc(FX), Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY) + FHeight) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX), Trunc(FY) + FHeight)) then
    if (FWorld.IsAvailablePos(Trunc(FX), Trunc(FY) - 1) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY) - 1 + FHeight) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY) - 1) and FWorld.IsAvailablePos(Trunc(FX), Trunc(FY) - 1 + FHeight)) then
    begin
      FY := FY - 1; //Izgleda sasavo, ali Stojadin se sad penje uzbrdo
      FFlags := FFlags - [flgRunning];
    end
    else
    begin
    //znaci ne valja X
      repeat
        if FX < FPriorX then
          FX := FX + 1
        else
          FX := FX - 1;
      until (FWorld.IsAvailablePos(Trunc(FX), Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY) + FHeight) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX), Trunc(FY) + FHeight));
      FVx := 0; //Postavi Vx na 0 da se ne bi igrac privremeno "zalepio" za objekat ako se zakuco maksimalnom brzinom i treba vremena da deluje trenje
    end;


//Izracunavanje Vy i Y...
  //Ako je dosao do vrha svog skoka
  if (flgJumping in FFlags) and (FVy <= 0) then
  begin
    FFlags := FFlags - [flgJumping];
    FVy := 0;
  end;

  //Utvrditi da li lik pada, tj. da li je u vazduhu
  //Ako pada onda deluje gravitacija
  if ((flgGround in FEnvFlags) and (flgFalling in FFlags)) or (flgWater in FEnvFlags) then
    FVy := FVy - FGravity;

  FPriorY := FY; //Zapamti stari dobri Y
  FY := FY - FVy;

  //Proveri novi Y
  if not (FWorld.IsAvailablePos(Trunc(FX), Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY) + FHeight) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX), Trunc(FY) + FHeight)) then
  //znaci ne valja Y
  begin
    repeat
      if FY < FPriorY then
        FY := FY + 1
      else
        FY := FY - 1;
    until (FWorld.IsAvailablePos(Trunc(FX), Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY) + FHeight) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY)) and FWorld.IsAvailablePos(Trunc(FX), Trunc(FY) + FHeight));
    FVy := 0; //Isto ko Vx = 0;
  end;
end;

function TCharacter.X: Integer;
begin
  Result := Trunc(FX);
end;

function TCharacter.Y: Integer;
begin
  Result := Trunc(FY);
end;

procedure TCharacter.ParamCheck;
begin
  FSpecialUltraMegaDirtyWaterJump := False;
  if FWorld.IsAvailablePos(Trunc(FX), Trunc(FY) + FHeight + 1) and FWorld.IsAvailablePos(Trunc(FX) + FWidth, Trunc(FY) + FHeight + 1) then
    FFlags := FFlags + [flgFalling]
  else
    FFlags := FFlags - [flgFalling];

  if flgGround in FEnvFlags then
  begin
    FGravity := GroundGravity;
    FFriction := GroundFriction;
    FVelIncrement := GroundVelIncrement;
    if (flgRunning in FFlags) and (Abs(FVx) > 0) then
    begin
      FMaxVel := GroundLargeMaxVel;
      FJumpImpulse := GroundLargeJumpImpulse;
    end
    else
    begin
      FMaxVel := GroundSmallMaxVel;
      FJumpImpulse := GroundSmallJumpImpulse;
    end;
    FTurnDelay := GroundTurnDelay;
  end;

  if flgWater in FEnvFlags then
  begin
    FGravity := WaterGravity;
    FFriction := WaterFriction;
    FVelIncrement := WaterVelIncrement;
    FMaxVel := WaterMaxVel;
    FTurnDelay := WaterTurnDelay;
    if not (flgJumping in FFlags) then
    begin
      if FVy > 3 then FVy := 3;
      if FVy < -2 then FVy := -2;
    end;
  end;

  if flgAir in FEnvFlags then
  begin
    FGravity := AirGravity;
    FFriction := AirFriction;
    FVelIncrement := AirVelIncrement;
    FMaxVel := AirMaxVel;
    FJumpImpulse := AirJumpImpulse;
    FTurnDelay := AirTurnDelay;
  end;

  SpecialTop := FWorld.IsSpecialPos(Trunc(FX), Trunc(FY));
  SpecialBtm := FWorld.IsSpecialPos(Trunc(FX), Trunc(FY) + FHeight);

  case SpecialTop of
    MapCodeWater:
      FEnvFlags := [flgWater];
    MapCodeAir:
    begin
      if SpecialBtm = MapCodeWater then
      begin
        if not (flgJumping in FFlags) then
          if FVy > 0 then FVy := 16 * FGravity;
        if Trunc(FVy) > -0.25 then
          FSpecialUltraMegaDirtyWaterJump := True;
      end
      else
        FEnvFlags := [flgGround];
    end;
  end;

end;

function TCharacter.GetFrameIndex: Byte;
var
  FrameSwitchCount: Byte;
begin
  FrameSwitchCount := 0;
  Result := 0;
  if flgGround in FEnvFlags then
  begin
    if flgRunning in FFlags then
      FrameSwitchCount := 5
    else
      FrameSwitchCount := 8;

    if flgJumping in FFlags then
      Result := FrameJumping
    else if flgFalling in FFlags then
      Result := FrameFalling
    else if not ((flgGoingLeft in FFlags) or (flgGoingRight in FFlags)) then
      Result := FrameStanding
    else if FFrameSwitch then
      Result := FrameGoing
    else
      Result := FrameStanding;
  end;

  if (flgAir in FEnvFlags) then
  begin
    if flgJumping in FFlags then
      FrameSwitchCount := 2
    else
      FrameSwitchCount := 12;

    if FFrameSwitch then
      Result := FrameFalling
    else
      Result := FrameGoing;
  end;

  if (flgWater in FEnvFlags) then
  begin
    if (flgGoingUp in FFlags) or (flgGoingDown in FFlags) or (flgGoingRight in FFlags) or (flgGoingLeft in FFlags) then
      FrameSwitchCount := 6
    else
      FrameSwitchCount := 12;

    if FFrameSwitch then
      Result := FrameFalling
    else
      Result := FrameGoing;
  end;

  Inc(FFrameCounter);
  if FFrameCounter >= FrameSwitchCount then
  begin
    FFrameSwitch := not FFrameSwitch;
    FFrameCounter := 0;
  end;

  if flgFacingRight in FFlags then
    Result := Result + 5;
end;

procedure TCharacter.GoDown;
begin
  if flgWater in FEnvFlags then
  begin
    if FVy > WaterTurnDelay then
      FVy := WaterTurnDelay;
    FVy := FVy - WaterVelIncrement;
    FFlags := FFlags + [flgGoingDown];
  end;
  if flgLadder in FEnvFlags then
    FY := FY + 0.25;
end;

procedure TCharacter.StopDown;
begin
  if flgWater in FEnvFlags then
    FFlags := FFlags - [flgGoingDown];
end;

function TCharacter.IsAlive: Boolean;
begin
  Result := not ((FWorld.IsSpecialPos(Trunc(FX), Trunc(FY)) = MapCodeDeath) or (FWorld.IsSpecialPos(Trunc(FX) + FWidth, Trunc(FY) + FHeight) = MapCodeDeath) or (FWorld.IsSpecialPos(Trunc(FX) + FWidth, Trunc(FY)) = MapCodeDeath) or (FWorld.IsSpecialPos(Trunc(FX), Trunc(FY) + FHeight) = MapCodeDeath));
end;

function TCharacter.Vx: Extended;
begin
  Result := FVx;
end;

function TCharacter.Vy: Extended;
begin
  Result := FVy;
end;

procedure TCharacter.GoUp;
begin
  if flgWater in FEnvFlags then
  begin
    FFlags := FFlags + [flgGoingUp];
    if FVy < -WaterTurnDelay then
      FVy := -WaterTurnDelay;
    FVy := FVy + WaterVelIncrement;
  end;
  if (SpecialTop = MapCodeLadder) or (SpecialBtm = MapCodeLadder) then
    FEnvFlags := [flgLadder];
    
  if flgLadder in FEnvFlags then
    FY := FY - 0.25;
end;

procedure TCharacter.StopUp;
begin

end;

end.

